# -*- coding: utf-8 -*-

import argparse
import time
import sys
from os import PathLike
from typing import final, Self, Any
from types import TracebackType

import serial  # type: ignore
import pyftdi.serialext  # type: ignore
from pyftdi.ftdi import Ftdi  # type: ignore
from pyftdi.usbtools import UsbDeviceDescriptor, UsbTools  # type: ignore

import jn51xx_flasher

__all__ = ["Tweliter"]


@final
class Tweliter:
    __ftdi: Ftdi
    __url: str
    __device: UsbDeviceDescriptor
    __debugging: bool

    def __init__(
        self,
        url: str | None = None,
        debugging: bool = False,
    ):
        UsbTools.flush_cache()  # for reconnecting

        self.__ftdi = Ftdi()
        self.__debugging = debugging

        # Set __url and __device
        if url is not None:
            # URL was specified
            self.__url = url
            try:
                self.__device = self.__ftdi.get_identifiers(url)[0]
            except Exception:
                self.close()
                raise IOError(f"There's no device matches URL {url}")
        else:
            # Set dynamically
            if not sys.stdin.isatty():
                self.close()
                raise EnvironmentError("There's no console. Specify URL.")

            # Get available devices
            devices = self.__ftdi.list_devices()

            if len(devices) <= 0:
                # No devices
                self.__ftdi.close()
                raise IOError("There's no devices.")
            elif len(devices) == 1:
                # One device
                self.__url = "ftdi:///1"
                self.__device = devices[0][0]
            else:
                # Two or more devices -> Ask for the user
                print("Detected multiple devices: ")
                for index, device in enumerate(devices):
                    devinfo = device[0]
                    sn = getattr(devinfo, "sn")
                    description = getattr(devinfo, "description")
                    print(f"[{index}] {description} ({sn})")
                while True:
                    try:
                        selected_index = int(input("Select device number to use: "))
                        if selected_index < 0 or selected_index >= len(devices):
                            raise ValueError("Invalid number")
                        break
                    except ValueError as e:
                        print(e)
                self.__device = devices[selected_index][0]
                self.__url = f'ftdi://::{getattr(self.__device, "sn")}/1'

        if self.__debugging:
            print(f"[Debug] Set URL {self.__url}")
            print(f"[Debug] Set Device {self.__device}")

        serial_no = getattr(self.__device, "sn")
        description = getattr(self.__device, "description")
        print(f"Using {description} ({serial_no})")

        self.__ftdi.open_from_url(self.__url)
        self.__ftdi.set_bitmode(0, Ftdi.BitMode.CBUS)

        if not self.__ftdi.has_cbus:
            # CBUS gpio are not initialized; Invalid device
            self.close()
            raise IOError("Device is invalid. Cannot use CBUS pins.")

    def __del__(self) -> None:
        self.close()

    def reopen(self) -> None:
        self.close()
        self.__ftdi.open_from_url(self.__url)
        if not self.__ftdi.has_cbus:
            # CBUS gpio are not initialized; Invalid device
            self.close()
            raise IOError("Device is invalid. Cannot use CBUS pins.")

    def __enter__(self) -> Self:
        """Enter the runtime context related to this object."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        """Exit the runtime context and close the connection."""
        self.close()
        return None

    def close(self) -> None:
        if self.__ftdi.is_connected:
            self.__ftdi.close()

    @property
    def url(self) -> str:
        return self.__url

    @property
    def ftdi(self) -> Ftdi:
        return self.__ftdi

    def enter_program_mode(self) -> None:
        self.__ftdi.set_cbus_direction(0b00001100, 0b00001100)
        self.__ftdi.set_cbus_gpio(0b00000100)  # Set PRG Low, RST High
        time.sleep(0.1)
        self.__ftdi.set_cbus_gpio(0b00000000)  # Set PRG Low, RST Low
        time.sleep(0.5)
        self.__ftdi.set_cbus_gpio(0b00000100)  # Set PRG Low, RST High
        time.sleep(0.2)

    def reset_device(self, set_low: bool = False) -> None:
        if not set_low:
            self.__ftdi.set_cbus_direction(0b00000100, 0b00000100)
            self.__ftdi.set_cbus_gpio(0b00000000)  # Set RST Low
            time.sleep(0.5)
            self.__ftdi.set_cbus_gpio(0b00000100)  # Set RST High
        else:
            self.__ftdi.set_cbus_direction(0b00000110, 0b00000110)
            self.__ftdi.set_cbus_gpio(0b00000100)  # Set RST High, SET Low
            time.sleep(0.1)
            self.__ftdi.set_cbus_gpio(0b00000000)  # Set RST Low, SET Low
            time.sleep(0.5)
            self.__ftdi.set_cbus_gpio(0b00000100)  # Set RST High, SET Low
        time.sleep(0.2)

    def enter_interactive_mode(self) -> None:
        self.reset_device(True)

    def get_serial_instance(self, **options: dict[str, Any]) -> serial.Serial:
        default_options = {
            "baudrate": 115200,
            "bytesize": serial.EIGHTBITS,
            "parity": serial.PARITY_NONE,
            "stopbits": serial.STOPBITS_ONE,
            "timeout": 2,
        }
        final_options = default_options | options
        ser: serial.Serial = pyftdi.serialext.serial_for_url(
            self.__url, **final_options
        )
        if ser.is_open:
            ser.close()
        if not ser.is_open:
            ser.open()
        ser.reset_output_buffer()
        ser.reset_input_buffer()
        return ser

    def get_startup_message_after(self, ser: serial.Serial, prefix: str) -> str:
        if not ser.is_open:
            ser.open()
        ser.baudrate = 115200
        ser.timeout = 1
        ser.reset_input_buffer()
        self.reset_device()
        ser.read_until(prefix.encode("utf-8"))
        line: bytes = ser.readline()
        try:
            message: str = line.decode("utf-8").strip()
        except UnicodeDecodeError:
            message = ""
        return message

    def write(
        self, ser: serial.Serial, file: PathLike[str] | str, verify: bool = False, retry: int = 2
    ) -> None:
        if not ser.is_open:
            ser.open()
        for i in range(1 + retry):
            self.enter_program_mode()
            try:
                self.write_firmware(ser, file, verify=verify)
                return
            except RuntimeError as e:
                print(e)
                if i < retry:
                    print("Retrying...")
                    continue
        raise RuntimeError("Failed to write")

    @staticmethod
    def write_firmware(
        ser: serial.Serial, file: PathLike[str] | str, verify: bool = False
    ) -> None:
        ser.reset_output_buffer()
        ser.reset_input_buffer()
        ser.timeout = 2
        ser.baudrate = 38400
        flasher = jn51xx_flasher.Flasher(ser, "none")  # type: ignore
        try:
            flasher.run("write", file)
            if verify:
                flasher.run("verify", file)
        except RuntimeError as e:
            raise RuntimeError(f"Failed to write ({e})")
        finally:
            ser.baudrate = 115200


def main() -> None:
    # Parse command line arguments
    parser = argparse.ArgumentParser(description="Write TWELITE BLUE/RED firmware")
    parser.add_argument(
        "--url", type=str, default=None, help="Device URL starting with ftdi://"
    )
    parser.add_argument(
        "--verify",
        action="store_true",
        help="Verify firmware after writing",
    )
    parser.add_argument(
        "--startmsg",
        type=str,
        default="!INF MONO WIRELESS",
        help="Prefix for startup message to check",
    )
    parser.add_argument(
        "--retry",
        type=int,
        default=2,
        help="Retry count in case of firmware writing failure",
    )
    parser.add_argument("file", help="Firmware file to write")
    args = parser.parse_args()

    try:
        with Tweliter(args.url) as liter:
            # Write
            ser = liter.get_serial_instance()
            try:
                liter.write(ser, args.file, verify=args.verify, retry=args.retry)
            except RuntimeError:
                print("Failed to write")
                return
            # Show startup message
            print(liter.get_startup_message_after(ser, args.startmsg))
    except IOError as e:
        print(e)


if __name__ == "__main__":
    main()
